/******************************************************************************
 * Coach.cc
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Coach
 *
 * requires:
 *
 *
 * Created on: 05.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors:
 *****************************************************************************/


#include "Coach.hh"
#include "PlayerType.hh"
#include "Communicator.hh"
#include "debug.hh"
#include "ConfigFile.hh"
#include "ParsedMsg.hh"
#include "Parser.hh"

#include "WorldModel.hh"

// ...#define DEBUG_*
// Please include any "#define DEBUG_*" lines in your
// "debugforuser.hh" file!


Coach::Coach( /*history, adviceQueue*/ ) {
  parser = new Parser();
  communicator = new Communicator();
  worldmodel = new WorldModel();
  isSynchronizedMode = false;
  sParam = NULL;
  pParam = NULL;
  playmode = PM_BeforeKickOff;
}


Coach::Coach (WorldModel *wm) {
	worldmodel = wm;
	communicator = new Communicator();
	parser = new Parser();
	isSynchronizedMode = false;
  sParam = NULL;
  pParam = NULL;
}


Coach::~Coach() {
  delete communicator;
  delete parser;
  delete worldmodel;
}



void Coach::init( const char *serverHost, int serverPort, const char *teamname ) {

  this->teamname = teamname;
  parser->setTeamName(teamname);
  char buf[65535];
  communicator->openConnection( serverHost, serverPort,
				teamname, ConfigFile::serverVersion,
				buf );

  ParsedMsg *parsedMsg = parser->parseMsg( (char *)buf );
  integrateMessage( parsedMsg );

  communicator->sendMessage( "(eye on)" );
}


void Coach::run() {
  char buf[65535];
  bool serverAlive = true;
  worldmodel->setPlayerTypes(&playerTypes);
  worldmodel->setPlaymode((Playmode)playmode);
	int lastcycle = 0;
  while ( serverAlive ) {
    serverAlive = communicator->waitForAndReadMessage( buf );
    if ( serverAlive ) {
      DEBUG2_PARSER( "\nCoach::run(): Message arrived: \n", buf );
      ParsedMsg *parsedMsg = parser->parseMsg( (char *)buf ); // new object!
      integrateMessage( parsedMsg ); // deletion of object!
      
      if (worldmodel->time != lastcycle) {
	worldmodel->saveState((Playmode)playmode);
      }
      lastcycle = worldmodel->time;
#if defined (DEBUG_WORLDMODEL)
      //worldmodel->printState();
#endif
    }
  }

  DEBUG1( "Disconnect: Server sent no message for 5 seconds." );
}

static bool sG_done = false;
static bool sG_rev  = false;
static bool sG_arrived[32][8] = {{false}};

static const char *red   = "(team_graphic (%u %u \"8 8 1 1\" \"R c #FF0000\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\"))";
static const char *white = "(team_graphic (%u %u \"8 8 1 1\" \"R c #FFFFFF\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\" \"RRRRRRRR\"))";
void Coach::sendGraphicPart(int x, int y){
  char msg[256];
  if (x > 1){
    if (!y%2)   sprintf(msg, red,   x, y);
    else        sprintf(msg, white, x, y);
  }else{
    if (y%2){
      if (!x%2) sprintf(msg, white, x, y);
      else      sprintf(msg, red,   x, y);    
    }else{
      if (!x%2) sprintf(msg, red,   x, y);
      else    	sprintf(msg, white, x, y);    
    }
  }
  //DEBUG2("Sending: ", msg);
  communicator->sendMessage(msg);
}

void Coach::sendGraphic(){
  // Dont send if already sent
  if (sG_done == true) return;
  // Dont send if we are not at the start of the game
  if (playmode != PM_BeforeKickOff) return;
  // Dont send if we are playing
  if (worldmodel->time > 1) return;

  // We will send the file from different sides of the image
  // This way we try to compensate lag. eg. send the end of the
  // image so we wait long enough for the confirmation of the
  // arrival of the first lines from the server.

  int sent = 0;
  sG_rev = !sG_rev;

  if (!sG_rev){
    for(int y = 0; y < 8; y++){
      for(int x = 0; x < 32; x++){
	if (sG_arrived[x][y]) continue;
	sendGraphicPart(x,y);
	sent++;
	if (sent == 5) return;
      }
    }
  }else{
    for (int y = 7; y >= 0; y--){
      for(int x = 31; x >= 0; x--){
	if (sG_arrived[x][y]) continue;
	sendGraphicPart(x,y);
	sent++;
	if (sent == 5) return;
      }
    }	
  }

  // we have sent no parts, all arrived...
  if (sent == 0){
    cout << " Team graphic sent... " << endl; 
    sG_done = true;
  }
}


static int numCyclesInPlayOn = 0;
static int lastSum = 0;
static int msgCount = 0;
static float lastDefLineX = 0;
static int lastFormation = 0;

void Coach::sendDetectedPlayerTypes(){
  if (!ConfigFile::usePlayerTypeDetection) return;
  int pT[11] = {0};
  int sum = 0;

  vector<Player>::iterator i;
  for (i  = worldmodel->otherTeam.begin();
       i != worldmodel->otherTeam.end(); ++i){

    pT[i->getUNum() - 1] = i->getFittestType();
    sum += pT[i->getUNum() - 1];
  }

  char msg[128];
  if (sum != lastSum && sum != 0) { // sending player types
    sprintf(msg, "(say (freeform \"(0 (1 0) (2 %i) (3 %i) (4 %i) (5 %i) (6 %i) (7 %i) (8 %i) (9 %i) (10 %i) (11 %i))\"))", pT[1], pT[2], pT[3], pT[4], pT[5], pT[6], pT[7], pT[8], pT[9], pT[10]);
    communicator->sendMessage(msg);
    msgCount++;
    lastSum = sum;
  }
}

void Coach::sendDetectedOffsideX(){
  char msg[128];
  // We dont need to send those informations if we dont detect those values ;)
  if (!ConfigFile::useOffsideXDetection) return;

  if (lastDefLineX != worldmodel->getOtherDefenseLine()) { // sending defense line x position
    memset (msg, '\0', sizeof(msg));
    sprintf (msg, "(say (freeform \"(1 (%0.4f))\"))", worldmodel->getOtherDefenseLine());
    communicator->sendMessage(msg);
    lastDefLineX = worldmodel->getOtherDefenseLine();
    msgCount++;
  }
}

void Coach::sendFormation() {
  char msg[128];
  // We dont need to send those informations if we dont detect those values ;)
  if (!ConfigFile::useFormationDetection || playmode == PM_PlayOn) return; 
    if (lastFormation != worldmodel->getFormation()) { // sending defense line x position
    memset (msg, '\0', sizeof(msg));
    sprintf (msg, "(say (freeform \"(2 (%i))\"))", worldmodel->getFormation());
    communicator->sendMessage(msg);
    lastFormation = worldmodel->getFormation();
    msgCount++;
  }
}

void Coach::sendDetectedValues(){
  // Are we allowed to send those messages?
  if (playmode == PM_PlayOn){
    if (numCyclesInPlayOn <= 600 ||
	numCyclesInPlayOn %  600 >= 20) return;
  }

  sendDetectedPlayerTypes();
  sendDetectedOffsideX();
  if (worldmodel->time >= 200 || worldmodel->time == 0) sendFormation();
}

// think function for test games (josh)
void Coach::testThink() {
  static int unum = 11;
  if ((serverTime == 0) &&
      (unum > 1) &&
      (pParam->player_types > 1) &&
      ((int)pParam->player_types == (int)playerTypes.size())) {
    vector<PlayerType>::iterator best = playerTypes.begin();
    vector<PlayerType>::iterator pt;
    for (pt = playerTypes.begin(); pt < playerTypes.end(); pt++) {
      if ((pt->aValue > best->aValue) &&
	  (((int)pt->used < (int)pParam->pt_max) || (pt->getTypeId() == 0))) {
	best = pt;
      }
    }
    char msg[30];
    if (best->getTypeId() != 0) {
      sprintf(msg, "(change_player_type %d %d)", unum, best->getTypeId());
      communicator->sendMessage(msg);
      best->used++;
    }
    unum--;
  }
}

void Coach::integrateMessage( ParsedMsg *parsedMsg ) {
  switch ( parsedMsg->getType() ) {

  case HearPlayer_Msg:{
    HearPlayerMsg *msg = (HearPlayerMsg *)parsedMsg;
    serverTime = msg->serverTime;
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    delete msg;
    break;
  }
    
  case HearReferee_Msg: {
    HearRefereeMsg *msg = (HearRefereeMsg *)parsedMsg;
    serverTime = msg->serverTime;
    playmode = msg->playmode;
    worldmodel->setPlaymode((Playmode)playmode);
    if (playmode == PM_PlayOn){
      numCyclesInPlayOn = 0;
    }
    //      printf("HearReferee: time %i, pm %i\n", serverTime, playmode);
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    delete msg;
    break;
  }

  case SeeGlobal_Msg:{
    SeeGlobalMsg *msg = (SeeGlobalMsg *)parsedMsg;
    serverTime = msg->serverTime;
    //if (!isSynchronizedMode) distributeHeterogenousPlayers();
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    worldmodel->integrateSeeGlobalMsg(msg);  
    delete msg;

    if (!isSynchronizedMode) {
      if (ConfigFile::useHeterogenousPlayerTypes)
	testThink();
      if (ConfigFile::usePlayerTypeDetection || 
	  ConfigFile::useOffsideXDetection || ConfigFile::useFormationDetection)
	sendDetectedValues();
      if (ConfigFile::sendTeamGraphic)
	sendGraphic();
      if (playmode == PM_PlayOn){
	numCyclesInPlayOn++;
      }
    }

    break;
  }

  case OkLook_Msg:{
    OkLookMsg *msg = (OkLookMsg *)parsedMsg;
    serverTime = msg->serverTime;
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    delete msg;
    break;
  }
    
  case OkSay_Msg:{
    OkSayMsg *msg = (OkSayMsg *)parsedMsg;
    DEBUG1_PARSER("OkSayMsg parsed");
    delete msg;
    break;
  }
    
  case OkTeamnames_Msg:{
    OkTeamnamesMsg *msg = (OkTeamnamesMsg *)parsedMsg;
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    delete msg;
    break;
  }
    
  case CLangVersPlayer_Msg:{
    CLangVersPlayerMsg *msg = (CLangVersPlayerMsg *)parsedMsg;
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    delete msg;
    break;
  }
    
  case InitSideOk_Msg: {
    InitSideOkMsg *msg = (InitSideOkMsg *)parsedMsg;
    DEBUG2_PARSER( "InitSideOkMsg parsed, fieldSide: ", 
		   msg->fieldSide );
    ourSide = msg->fieldSide;
    delete msg;
    break;
  }
    
  case OkEye_Msg:{
    OkEyeMsg *msg = (OkEyeMsg *)parsedMsg;
    if ( msg->eyeOn ) {
      DEBUG1_PARSER( "EyeOkMsg parsed: eye on" );
    } else {
      DEBUG1_PARSER( "EyeOkMsg parsed: eye off" );
    }
    delete msg;
    break;
  }
    
  case OkChangePlayerType_Msg:{
    OkChangePlayerTypeMsg *msg = (OkChangePlayerTypeMsg *)parsedMsg;
    DEBUG4_PARSER( "OkChangePlayerTypeMsg parsed, uNum: ",
		   msg->uNum, "  playerType: ", msg->playerType );
    delete msg;
    break;
  }
  
  case ChangePlayerType_Msg: {
    ChangePlayerTypeMsg *msg = (ChangePlayerTypeMsg *)parsedMsg;
    DEBUG2_PARSER( "ChangePlayerTypeMsg parsed, uNum; ", msg->uNum);
    worldmodel->integrateChangePlayerTypeMsg(msg);
    delete msg;
    break; 
  }
    
  case Think_Msg:{
    DEBUG1_PARSER( "ThinkMsg parsed" );
    isSynchronizedMode = true;
    testThink();
    sendDetectedValues();
    sendGraphic();
    communicator->sendMessage("(done)");
    delete parsedMsg;
    break;
  }

  case W_NoTeamFound:{
    DEBUG1( "Warning: Coach::integrateMessage: WNoTeamFoundMsg parsed" );
    delete (WNoTeamFoundMsg *)parsedMsg;
    break;
  }
    
  case W_NoSuchPlayer:{
    DEBUG1( "Warning: Coach::integrateMessage: WNoSuchPlayerMsg parsed" );
    delete (WNoSuchPlayerMsg *)parsedMsg;
    break;
  }
    
  case W_CannotSubWhilePlayOn:{
    DEBUG1( "Warning: Coach::integrateMessage: WCannotSubWhilePlayOnMsg parsed" );
    delete (WCannotSubWhilePlayOnMsg *)parsedMsg;
    break;
  }
    
  case W_NoSubsLeft:{
    DEBUG1( "Warning: Coach::integrateMessage: WNoSubsLeftMsg parsed" );
    delete (WNoSubsLeftMsg *)parsedMsg;
    break;
  }
    
  case W_MaxOfThatPlayerTypeOnField:{
    DEBUG1( "Warning: Coach::integrateMessage: WMaxOfThatPlayerTypeOnFieldMsg parsed" );
    delete (WMaxOfThatPlayerTypeOnFieldMsg *)parsedMsg;
    break;
  }
    
  case W_CannotChangeGoalie:{
    DEBUG1( "Warning: Coach::integrateMessage: WCannotChangeGoalieMsg parsed" );
    delete (WCannotChangeGoalieMsg *)parsedMsg;
    break;
  }
    
  case PlayerType_Msg:{
    PlayerTypeMsg *msg = (PlayerTypeMsg *)parsedMsg;
    PlayerType pt = PlayerType();
    DEBUG1( "PlayerTypeMsg parsed: ( id" <<
	    msg->id << ",\n\tspeed_max " << msg->player_speed_max << ",\n\tstamina_inc " <<
	    msg->stamina_inc_max << ",\n\tdecay " << msg->player_decay << ",\n\tinertia_moment " <<
	    msg->inertia_moment << ",\n\tdash_power_rate " << msg->dash_power_rate << ",\n\tsize " <<
	    msg->player_size << ",\n\tkickable_margin " << msg->kickable_margin << ",\n\tkick_rand " <<
	    msg->kick_rand << ",\n\textra_stamina " << msg->extra_stamina << ",\n\teffort_max " <<
	    msg->effort_max << ",\n\teffort_min " << msg->effort_min << " )" );
    pt.setAll(msg->id,
	      msg->player_speed_max,
	      msg->stamina_inc_max,
	      msg->player_decay,
	      msg->inertia_moment,
	      msg->dash_power_rate,
	      msg->player_size,
	      msg->kickable_margin,
	      msg->kick_rand,
	      msg->extra_stamina,
	      msg->effort_max,
	      msg->effort_min);

    float aValue;

    //Josh Mode
    if(ConfigFile::useHPT_Mode == 0){
      float diff_param[4] = { 107.0-0.0, 2.0-5.0, 18.0-1.0, 0.0-34.0 };
      //unused:
      //  float rel_param[4] = { 107.0/1.0, 2.0/5.0, 18.0/1.0, 1.0/34.0 }; // 0->1 for division
      aValue = (( msg->player_decay    - 0.4   ) / 0.2   ) * diff_param[0]
	+      (( msg->dash_power_rate - 0.006 ) / 0.002 ) * diff_param[1]
	+      (( msg->kickable_margin - 0.7   ) / 0.2   ) * diff_param[2]
        +      (( msg->extra_stamina   - 0.0   ) / 100.0 ) * diff_param[3];
    }
    
    //Cord Mode
    if(ConfigFile::useHPT_Mode == 1){
      float highSpeed        = (msg->dash_power_rate * msg->effort_max * 100) / 
	                       (1.0 - msg->player_decay);
      float highSpeedPoints  = min( (highSpeed - 1.0) * 500.0, 100.0);
      float staminaIncPoints;
      if(msg->stamina_inc_max >= 30.0){
	  if(msg->stamina_inc_max >= 35.0){
	      staminaIncPoints = (msg->stamina_inc_max  - 25.0) * 5.0;
	  }else{
	      staminaIncPoints = (msg->stamina_inc_max  - 30.0) * 10.0;
	  }
      }else{
	  staminaIncPoints = -100.0f; 
      }
      aValue                 = highSpeedPoints + staminaIncPoints; 
    }

    pt.aValue = aValue;
    pt.used   = 0;
    playerTypes.push_back(pt);
    delete msg;
    break;
  }
    
  case PlayerParam_Msg:{
    PlayerParamMsg *msg = (PlayerParamMsg *)parsedMsg;
    differentPlayerTypes = msg->player_types;
    maximumSubstitutions = msg->subs_max;
#if defined (DEBUG_PARSER)
    msg->print();
#endif
    if (pParam != NULL) delete pParam;
    pParam = msg;
    break;
  }
    
  case ServerParam_Msg:{
    ServerParamMsg *msg = (ServerParamMsg *)parsedMsg;
    DEBUG1_PARSER( "ServerParamMsg parsed: many, many values..." );
    if (sParam != NULL) delete sParam;
    sParam = msg;
    break;
  }

  case TeamGraphicOk_Msg:{
    TeamGraphicOkMsg *msg = (TeamGraphicOkMsg *)parsedMsg;
    sG_arrived[msg->x][msg->y] = true;
    break;
  }
    
  case E_NoMoreTeamOrPlayerOrGoalie:{
    DEBUG1( "Error: Coach::integrateMessage: ENoMoreTeamOrPlayerOrGoalieMsg parsed" );
    delete (ENoMoreTeamOrPlayerOrGoalieMsg *)parsedMsg;
    break;
  }

  case E_NoSuchTeamOrAlreadyHaveCoach:{
    DEBUG1( "Error: Coach::integrateMessage: ENoSuchTeamOrAlreadyHaveCoachMsg parsed" );
    delete (ENoSuchTeamOrAlreadyHaveCoachMsg *)parsedMsg;
    break;
  }
  
  case E_UnknownCommand:{
    DEBUG1( "Error: Coach::integrateMessage: EUnknownCommandMsg parsed" );
    delete (EUnknownCommandMsg *)parsedMsg;
    break;
  }

  case E_IllegalCommandForm:{
    DEBUG1( "Error: Coach::integrateMessage: EIllegalCommandFormMsg parsed" );
    delete (EIllegalCommandFormMsg *)parsedMsg;
    break;
  }

  case E_OutOfRangePlayerType:{
    DEBUG1( "Error: Coach::integrateMessage: EOutOfRangePlayerTypeMsg parsed" );
    delete (EOutOfRangePlayerTypeMsg *)parsedMsg;
    break;
  }

  case E_IllegalMode:{
    DEBUG1( "Error: Coach::integrateMessage: EIllegalModeMsg parsed" );
    delete (EIllegalModeMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyFreeformMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyFreeformMessagesMsg parsed" );
    delete (ESaidTooManyFreeformMessagesMsg *)parsedMsg;
    break;
  }

  case E_CannotSayFreeformWhilePlayon:{
    DEBUG1( "Error: Coach::integrateMessage: ECannotSayFreeformWhilePlayonMsg parsed" );
    delete (ECannotSayFreeformWhilePlayonMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyMetaMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyMetaMessagesMsg parsed" );
    delete (ESaidTooManyMetaMessagesMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyAdviceMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyAdviceMessagesMsg parsed" );
    delete (ESaidTooManyAdviceMessagesMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyInfoMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyInfoMessagesMsg parsed" );
    delete (ESaidTooManyInfoMessagesMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyDefineMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyDefineMessagesMsg parsed" );
    delete (ESaidTooManyDefineMessagesMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyDelMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyDelMessagesMsg parsed" );
    delete (ESaidTooManyDelMessagesMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyRuleMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyRuleMessagesMsg parsed" );
    delete (ESaidTooManyRuleMessagesMsg *)parsedMsg;
    break;
  }
  
  case E_CouldNotParseSay:{
    DEBUG1( "Error: Coach::integrateMessage: ECouldNotParseSayMsg parsed" );
    delete (ECouldNotParseSayMsg *)parsedMsg;
    break;
  }

  case E_SaidTooManyMessages:{
    DEBUG1( "Error: Coach::integrateMessage: ESaidTooManyMessagesMsg parsed" );
    delete (ESaidTooManyMessagesMsg *)parsedMsg;
    break;
  }

  case E_UnknownError:{
    EUnknownErrorMsg *msg = (EUnknownErrorMsg *)parsedMsg;
    DEBUG2( "Error: Coach::integrateMessage: EUnknownErrorMsg parsed: ",
	    msg->errorMsg );
    delete msg;
    break;
  }
    
  default:{
    DEBUG1("ERROR: Coach::integrateMessage: unknown Message-Type.");
    delete parsedMsg;
  }
  }
}
