/*******************************************************************************
 * common.hh
 * Copyright (c) 2003 Virtual Werder
 *
 * implements:
 *   some constants, needed by many components
 *   enum FieldSide
 *   enum ParsedMsgType
 *   enum MarkerId
 *   enum Playmode
 *   enum PenaltySide
 *   enum Command
 *   enum ViewQuality
 *   enum ViewMode
 *   function sign()
 *
 * requires:
 *
 * Created on: 07.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors: 
 *
 ******************************************************************************/
#ifndef COMMON_HH
#define COMMON_HH


/** @addtogroup Main
 *  
 *  @{
 */
 
#include <limits.h>
#include <math.h>
#include <float.h>


const int    BIG_INT    = INT_MAX;
const float  BIG_FLOAT  = FLT_MAX;
const double BIG_DOUBLE = DBL_MAX;

const int    UNDEFINED_INT    = BIG_INT;
const float  UNDEFINED_FLOAT  = 999.333;
const double UNDEFINED_DOUBLE = BIG_DOUBLE;

const int TIME_UNDEFINED = -1;

const double PI = M_PI;

/** Value used for floating point equality tests. */
const float  EPSILON  = 0.001;

bool isValidAngle(const float ang);
bool isValidBool(const bool b);
bool isUndefined(const float f);
bool isUndefined(const int i);

enum MarkerId {
  //goals
  GOAL_L  = 4,
  GOAL_R  = 5,

  //center flags
  FLAG_C  = 6,
  FLAG_CT = 62,
  FLAG_CB = 63,

  //corner flags
  FLAG_LT = 42,
  FLAG_LB = 43,
  FLAG_RT = 52,
  FLAG_RB = 53,

  //penalty area flags
  FLAG_PLT = 742,
  FLAG_PLC = 746,
  FLAG_PLB = 743,
  FLAG_PRT = 752,
  FLAG_PRC = 756,
  FLAG_PRB = 753,

  //jambs
  FLAG_GLT = 842,
  FLAG_GLB = 843,
  FLAG_GRT = 852,
  FLAG_GRB = 853,

  //top flags
  FLAG_TL50 = 2450,
  FLAG_TL40 = 2440,
  FLAG_TL30 = 2430,
  FLAG_TL20 = 2420,
  FLAG_TL10 = 2410,
  FLAG_T0   = 200,
  FLAG_TR10 = 2510,
  FLAG_TR20 = 2520,
  FLAG_TR30 = 2530,
  FLAG_TR40 = 2540,
  FLAG_TR50 = 2550,

  //bottom flags
  FLAG_BL50 = 3450,
  FLAG_BL40 = 3440,
  FLAG_BL30 = 3430,
  FLAG_BL20 = 3420,
  FLAG_BL10 = 3410,
  FLAG_B0   = 300,
  FLAG_BR10 = 3510,
  FLAG_BR20 = 3520,
  FLAG_BR30 = 3530,
  FLAG_BR40 = 3540,
  FLAG_BR50 = 3550,

  //left flags
  FLAG_LT30 = 4230,
  FLAG_LT20 = 4220,
  FLAG_LT10 = 4210,
  FLAG_L0   = 400,
  FLAG_LB10 = 4310,
  FLAG_LB20 = 4320,
  FLAG_LB30 = 4330,

  //right flags
  FLAG_RT30 = 5230,
  FLAG_RT20 = 5220,
  FLAG_RT10 = 5210,
  FLAG_R0   = 500,
  FLAG_RB10 = 5310,
  FLAG_RB20 = 5320,
  FLAG_RB30 = 5330,

  MARKER_UNKNOWN = 5555
};


/**
 * Contains all possible play modes of the simulation league soccer game. For
 * better usability there is a subdivision in Our<playmode> and Their<playmode>
 * not in Left<playmode> and Right<playmode>.
 */
enum Playmode {
  //missing here, present in ss: PM_Null
  PM_BeforeKickOff,            //0
  PM_TimeOver,                 //1
  PM_PlayOn,                   //2
  PM_OurKickOff,               //3    //unser Anstoss in der Mitte
  PM_TheirKickOff,             //4    //deren Anstoss in der Mitte

  PM_OurKickIn,                //5    //unser Einwurf
  PM_TheirKickIn,              //6    //deren Einwurf

  PM_OurFreeKick,              //7    //<=> also if goalie caught the ball
  PM_TheirFreeKick,            //8

  PM_OurCornerKick,            //9
  PM_TheirCornerKick,          //10

  PM_OurGoalKick,              //11   //unser Abstoss vom 5 Meter-Raum-Eck
  PM_TheirGoalKick,            //12   //deren Abstoss vom 5 Meter-Raum-Eck

  PM_OurGoal,                  //13   //<=> we scored a goal
  PM_TheirGoal,                //14   //<=> they scored a goal

  PM_DropBall,                 //15
  PM_OurOffside,               //16   //<=> we ran offside, followed by TheirFreeKick
  PM_TheirOffside,             //17   //<=> they ran offside, followed by OurFreeKick


  PM_OurPenaltyKick,           //18
  PM_TheirPenaltyKick,         //19

  //missing here in ss present: PM_FirstHalfOver
  PM_Pause,                    //20
  PM_HumanJudge,               //21   //human interception (???)

  //the following 2 are missing in ss:
  PM_OurFoul,                  //22
  PM_TheirFoul,                //23

  PM_OurFoulCharge,            //24
  PM_TheirFoulCharge,          //25
  PM_OurFoulPush,              //26
  PM_TheirFoulPush,            //27
  PM_OurFoulMultipleAttack,    //28
  PM_TheirFoulMultipleAttack,  //29
  PM_OurFoulBallout,           //30
  PM_TheirFoulBallout,         //31
  PM_OurBackPass,              //32
  PM_TheirBackPass,            //33
  PM_OurFreeKickFault,         //34
  PM_TheirFreeKickFault,       //35

  //the following 4 are missing in ss
  PM_OurGoalieCatchBall,       //36
  PM_TheirGoalieCatchBall,     //37
  PM_HalfTime,                 //38
  PM_TimeUp,                   //39

  PM_OurIndirectFreeKick,      //40
  PM_TheirIndirectFreeKick,    //41
  PM_OurPenaltySetup,
  PM_TheirPenaltySetup,
  PM_OurPenaltyReady,           
  PM_TheirPenaltyReady,        //45
  PM_OurPenaltyTaken,
  PM_TheirPenaltyTaken,
  PM_OurPenaltyMiss,
  PM_TheirPenaltyMiss,
  PM_OurPenaltyScore,          //50
  PM_TheirPenaltyScore,

  PM_OurCatchFault,
  PM_TheirCatchFault,
  //missing here in ss present: PM_MAX
  PM_UNKNOWN
};

enum CoFormation {
  FO_Default,
  FO_343,
  FO_442,
  FO_352,
  FO_532,
  FO_433,
  FO_424,
  FO_334,
  FO_325,
  FO_Unknown
};

enum CoTactic {
  TA_Default,
  TA_Offensive,
  TA_Defensive,
  TA_Controlled,
  TA_Reserved,
  TA_Crowbar,
  TA_Stronghold
};

enum BehaviourType{
  B_Goalie = 1,
  B_WingDefenderLeft,
  B_CenterBackDefenderLeft,
  B_CenterBackDefenderRight,
  B_WingDefenderRight,
  B_WingMidfielderLeft,
  B_CenterMidfielder,
  B_WingMidfielderRight,
  B_WingAttackerLeft,
  B_CenterAttacker,
  B_WingAttackerRight
};


enum ParsedMsgType {
  Unknown_Msg = -1,
  InitSideOk_Msg = 0,
  OkSay_Msg,
  OkChangePlayerType_Msg,
  ChangePlayerType_Msg,
  OkLook_Msg,
  OkEye_Msg,
  OkTeamnames_Msg,
  ServerParam_Msg,
  PlayerParam_Msg,
  PlayerType_Msg,
  SeeGlobal_Msg,
  HearReferee_Msg,
  HearPlayer_Msg,
  CLangVersPlayer_Msg,

  TeamGraphicOk_Msg,

  Think_Msg,

  W_NoTeamFound = 1000,
  W_NoSuchPlayer,
  W_CannotSubWhilePlayOn,
  W_NoSubsLeft,
  W_MaxOfThatPlayerTypeOnField,
  W_CannotChangeGoalie,
  W_UnknownWarning,

  E_NoMoreTeamOrPlayerOrGoalie = 2000,
  E_NoSuchTeamOrAlreadyHaveCoach,
  E_UnknownCommand,
  E_IllegalCommandForm,
  E_OutOfRangePlayerType,
  E_IllegalMode,
  E_IllegalObjectForm,
  E_SaidTooManyFreeformMessages,
  E_CannotSayFreeformWhilePlayon,
  E_SaidTooManyMetaMessages,
  E_SaidTooManyAdviceMessages,
  E_SaidTooManyInfoMessages,
  E_SaidTooManyDefineMessages,
  E_SaidTooManyDelMessages,
  E_SaidTooManyRuleMessages,
  E_CouldNotParseSay,
  E_SaidTooManyMessages,
  E_UnknownError
};


enum PenaltySide{
  PS_Unknown, PS_OurGoal, PS_TheirGoal
};

/** This type is used for storage the side, on which we are playing */
enum FieldSide{
  FS_Left, FS_Right, FS_Unknown
};

/** is used for storage the type of soccerserver command */
enum Command{
  CMD_NONE,
  CMD_MOVE,
  CMD_KICK,
  CMD_DASH,
  CMD_TURN,
  CMD_GOALIE_CATCH,
  CMD_TURN_NECK,
  CMD_CHANGE_VIEW,
  CMD_SCORE,
  CMD_SAY,
  // added 11-21-02
  CMD_TACKLE,
  CMD_ATTENTION_TO,
  CMD_POINT_TO
};

/** A type for representing the view quality of a player. */
enum ViewQuality { VQ_HIGH, VQ_LOW, VQ_UNKNOWN };

/** A type for representing the view width of a player. */
enum ViewWidth   { VW_NARROW, VW_NORMAL, VW_WIDE, VW_UNKNOWN };

/**  */
enum LineId { L_Top, L_Bottom, L_Left, L_Right, L_Unknown };

/** */
enum PlayerSet {PS_Players, PS_Teammates, PS_Opponents, PS_Unknowns};


/**
 * This function lacks in algobase.h. It calculates the sign of a number.
 * @param the number, whose sign should be calculated
 * @return 1 if positive or equal to 0, -1 otherwise
 */
template <class Type > inline const int sign(const Type &x){
  return ( x >= 0 ? 1 : -1);
}

/** @} */

#endif //COMMON_HH

