/******************************************************************************
 * Player.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Player
 *
 * requires:
 *
 *
 * Created on: 21.01.2004 by aw
 * Maintainer: aw, mangels
 * Group:
 * Co-authors:
 *
 * $Id: Player.hh,v 1.4 2004/06/24 10:19:05 aw Exp $
 *****************************************************************************/


#ifndef PLAYER_HH
#define PLAYER_HH

#include "common.hh"
#include "Geometry.hh"
#include "PlayerType.hh"

using namespace std;

/**
   This class represents a player of the coach's worldmodel.
   @ingroup WorldModel
   @brief class Player in group WorldModel
*/
class Player {
private:
  /**
     player's position, represented as Vector2D
  */
  Vector2D position;

  /**
     player's velocity, represented as Vector2D
  */
  Vector2D velocity;

  /**
     player's uniform number, default 0
  */
  int uNum;

  /**
     true, if goalkeeper, default false
  */
  bool goalie;

  /**
     body direction
  */
  float bodyDir;

  /**
     neck direction
  */
  float neckDir;

  /**
     pointer to the parsed player types
  */
  PlayerType *type;
  
  /**
    average position values
  */
  float avgX;
  float avgY;
  
  /**
    number of used cycles
  */
  int avgCycles;
  
  /**
   player's type has been changed
  */
  bool typeChanged;
   

public:
  /**
     array which states which playerType is still possible
  */
  //float playerTypePossible[7];

  /**
     playerType of the agent we have is not are is this richtig
  */
  int playerType;

private:
  /******************************************/
  /** CALCULATED VALUES *********************/
  /******************************************/

  /**
     calculated top speed
  */
  float topSpeed;

  /**
     calculated kickable margin
  */
  float kickableMargin;

  /**
     inertia
  */
  float inertia;
  /**
      accel
  */
  float accel;
  /**
      decay
  */
  float decay;
  /**
      decayNum
  */
  float decayNum;

  /**
   * values for each player type calculated by the fitness function
   */
  float ptMatchingValues[7];

  /**
   * from the worldmodel predicted type
   */
  int predictedPlayerType;

public:

  /**
     construct a new player with uniform number 0
  */
  Player();

  /**
     constructs a new player with a given uniform number
     @param uNum player's uniform number
  */
  Player(const int uNum);

  /**
     destructor
     @todo to be implemented
  */
  ~Player();

  /**
   * set the player position
   * @param pos player position represented as Vector2D
   */
  void setPosition (const Vector2D pos);

  /**
   * set the player velocity
   * @param vel player velocity represented as Vector2D
   */
  void setVelocity (const Vector2D vel);

  /**
   * set the player position
   * @param posX x-coordinate
   * @param posY y-coordinate
   */
  void setPosition (const float posX, const float posY);

  /**
   * set the player velocity
   * @param posX x-velocity
   * @param posY y-velocity
   */
  void setVelocity (const float velX, const float velY);

  /**
   * set the body direction
   * @param dir body direction value
   */
  void setBodyDir (const float dir);

  /**
   * set the neck direction
   * @param dir neck direction value
   */
  void setNeckDir (const float dir);

  /**
   * set the player type
   * @param t player's type
   */
  void setType (PlayerType *t);

  /**
   * set the uniform number
   * @param uNum player's uniform number
   */
  void setUNum (const int uNum);

  /**
   * sets the goalkeeper flag
   * @param isGoalie true if player is goalkeeper
   */
  void setGoalie (const bool isGoalie);

  /**
   * set player's top speed
   * @param topSpeed calculated top speed
   */
  void setTopSpeed (const float topSpeed);

  /**
   * set player's kickable margin
   * @param kickableMargin calculated kickable margin
   */
  void setKickableMargin (const float km);

  /**
   * sets the predicted player type
   * @param predictedTypeId
   */
  void setPredictedPlayerType (int i);

  
  /**
   * sets whether the player's type changed
   * (otherTeam!!!)
   * @param bool changed
   */
  void setHasTypeChanged (bool changed);

  /**
   * updates the average X,Y position value
   */
  void updateAvgValues();
    
  /**
   * calculates the average x position
   * @return float avgX
   */
  float getAvgX(void) const {
    if (avgCycles == 0) return 0;
    return (float) (avgX / avgCycles);
  }
  
  /**
   * calculates the average y position
   * @return float avgy
   */
  float getAvgY(void) const {
    if (avgCycles == 0) return 0;
    return (float) (avgY / avgCycles);
  }
   
  /**
   * get the uniform number
   * @return int player's uniform number
   */
  int getUNum(void) const { return uNum; }

  /**
   * get the position as Vector2D
   * @return Vector2D player's position
   */
  Vector2D getPosition(void) const { return position; }

  /**
   * get the x coordinate of the player's posisition
   * @return float x-coordinate
   */
  float getPosX(void) const { return position.getX(); }

  /**
   * get the y coordinate of the player's posisition
   * @return float y-coordinate
   */
  float getPosY(void) const { return position.getY(); }

  /**
   * get the x value of the player's velocity
   * @return float velocity x-value
   */
  float getVelX(void) const { return velocity.getX(); }

  /**
   * get the y value of the player's velocity
   * @return float velocity y-value
   */
  float getVelY(void) const { return velocity.getY(); }

  /**
   * get the length of the vector of the player's velocity
   * @return float velocity
   */
  float getVelocity(void) const { return velocity.getLength(); }

  /**
   * get the body direction of the player
   * @return float body direction
   */
  float getBodyDir(void) const { return bodyDir; }

  /**
   * get the neck direction of the player
   * @return float neck direction
   */
  float getNeckDir(void) const { return neckDir; }

  /**
   * determines whether the player is goalkeeper or not
   * @return bool true, if player is goalkeeper
   */
  bool isGoalie(void) const { return goalie; }

  /**
   * @return bool true, if the player's type was changed, else false
   */
  bool hasTypeChanged(void) const { return typeChanged; }
  
  /**
   * get the calculated top speed
   * @return float top speed
   */
  float getTopSpeed(void) const { return topSpeed; }

  /**
   * get the calculated kickable margin
   * @return float kickable margin
   */
  float getKickableMargin(void) { return kickableMargin; }

  /**
   * get playerTypePossible
   * @ return playerTypePossible or not
   */
  //float getPlayerTypePossible(int i) { return playerTypePossible[i]; }

  float getInertia() { return inertia; }
  void setInertia(float f){ inertia = f; }
  float getAccel() { return accel; }
  void setAccel(float f){ accel = f; }
  float getDecay() { return decay; }
  void setDecay(float f){ decay = f; decayNum++; }
  float getDecayNum() { return decayNum; }
  // bool checkIfDoneWithPlayerTypes();

  /**
   * get the player type
   * @return PlayerType player's type
   */
  PlayerType* getType(void) const { return type; }

  /**
   * Fitness function, to determine which is the most fitting playertype
   */
  float getPlayerTypeMatching (int ptId,
                               float ptSpeedMax,
                               float ptKickableMarin,
                               float ptInertia, float ptAccel,
                               float ptDecay);

  /**
   * returns the matching value for a given player type
   * @param int player type id
   * @return int matchingValue
   */
  float getMatchingValue (int ptId) {
    if (ptId < 7 && ptId >= 0)
      return ptMatchingValues[ptId];
    else return 0;
  }
  
  /**
   * resets the player type prediction
   */
  void resetPlayerTypePrediction () {
    predictedPlayerType = 0;
    for (int i = 0; i < 7; i++) {
      ptMatchingValues[i] = 0;
    }
  }

  /**
   * determines wether a given type is the fittest, for this player
   * @param int player type id
   * @return true, if the given type is the fittest
   */
  bool isFittestType (int ptId);

  /**
   * @return int the fittest (the type with the highest matching value) type
   */
  int getFittestType ();

  /**
   * Method to change the matching value of a player type
   * @param int player type id
   * @param float new matching value
   */
  void setMatchingValue (int ptId, float matchingValue) {
    if (ptId < 7 && ptId >= 0)
      ptMatchingValues[ptId] = matchingValue;
  }
};

#endif // PLAYER_HH
